# pyinfra
# File: docs/conf.py
# Desc: minimal Sphinx config

# import os

from better import better_theme_path

from pyinfra import __version__


extensions = [
    # Official
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
]

source_suffix = '.rst'
master_doc = 'index'
project = 'pyinfra'
copyright = '2017, Nick Barrett'
author = 'Fizzadar'
version = 'develop'
pygments_style = 'sphinx'

# Theme style override
html_title = 'pyinfra {0}'.format(__version__)
html_short_title = 'Home'
html_theme = 'better'
html_theme_path = [better_theme_path]
html_static_path = ['static']

templates_path = ['templates']

html_sidebars = {
    '**': ['pyinfra_sidebar.html'],
}

html_theme_options = {
    'cssfiles': ['_static/pyinfra.css'],
    'scriptfiles': ['_static/sidebar.js'],
}



# pyinfra
# File: example/inventories/test.py
# Desc: basic IP based inventory for testing the example
#       see ./hostnames.py for a more advanced example

# Defines a group - group names must be defined in ALL_CAPS
linux = [
    # Ubuntu 14
    '20.20.20.21',
    # Ubuntu 15
    '20.20.20.26',
    # CentOS 6
    '20.20.20.22',
    # CentOS 7
    '20.20.20.23',
    # Debian 7
    '20.20.20.24',
    # Debian 8
    '20.20.20.27',
    # Fedora 23
    '20.20.20.28'
]

bsd = [
    # OpenBSD 5.8
    '20.20.20.25'
]



#!/usr/bin/env python

# pyinfra
# File: scripts/generate_empty_tests.py
# Desc: generate empty test files

from __future__ import print_function

import sys

from os import makedirs, path

from pyinfra.api.facts import get_fact_names
from pyinfra.api.operation import get_operation_names

WRITE_EMPTIES = '--write' in sys.argv
WRITE_LIMIT = [
    item for item in sys.argv[1:]
    if item != '--write'
]

EMPTY_FACT = '''{
    "output": [],
    "fact": null
}
'''

EMPTY_OPERATION = '''{
    "args": [],
    "kwargs": {},
    "facts": {},
    "commands": [],
    "exception": {}
}
'''


print('--> {0} empty fact tests'.format(
    'Generating' if WRITE_EMPTIES else 'Printing',
))
for fact in get_fact_names():
    if WRITE_LIMIT and fact not in WRITE_LIMIT:
        continue

    fact_folder = path.join('tests', 'facts', fact)

    if not path.exists(fact_folder):
        fact_file = path.join(fact_folder, 'empty.json')

        if WRITE_EMPTIES:
            makedirs(fact_folder)
            f = open(fact_file, 'w')
            f.write(EMPTY_FACT)
            f.close()

        print('    {0}: {1}'.format(
            'written' if WRITE_EMPTIES else 'missing',
            fact_file,
        ))


print()
print('--> {0} empty operation tests'.format(
    'Generating' if WRITE_EMPTIES else 'Printing',
))
for operation in get_operation_names():
    if WRITE_LIMIT and operation not in WRITE_LIMIT:
        continue

    operation_folder = path.join('tests', 'operations', operation)
    if not path.exists(operation_folder):
        operation_file = path.join(operation_folder, 'empty.json')

        if WRITE_EMPTIES:
            makedirs(operation_folder)
            f = open(operation_file, 'w')
            f.write(EMPTY_OPERATION)
            f.close()

        print('    {0}: {1}'.format(
            'written' if WRITE_EMPTIES else 'missing',
            operation_file,
        ))

print()
print('--> Done!')



#!/usr/bin/env python

# pyinfra
# File: docs/generate_module_docs.py
# Desc: generate rst docs from the module modules

import re
from importlib import import_module
from inspect import getargspec, getmembers
from types import FunctionType

from six.moves import range

from pyinfra import modules

MODULE_DEF_LINE_MAX = 90


def _title_line(char, string):
    return ''.join(char for _ in range(0, len(string)))


def _format_doc_line(line):
    # Bold the <arg>: part of each line
    line = re.sub(r'\+ ([a-z_\/]+)(.*)', r'+ **\1**\2', line)

    # Strip the first 4 characters (first indent from docstring)
    return line[4:]


def build_facts():
    for module_name in modules.__all__:
        lines = []

        print('--> Doing module: {0}'.format(module_name))
        module = import_module('pyinfra.modules.{0}'.format(module_name))

        lines.append(module_name.title())
        lines.append(_title_line('-', module_name))
        lines.append('')

        if module.__doc__:
            lines.append(module.__doc__)

        operation_functions = [
            (key, value._pyinfra_op)
            for key, value in getmembers(module)
            if (
                isinstance(value, FunctionType)
                and value.__module__ == module.__name__
                and getattr(value, '_pyinfra_op', False)
                and not value.__name__.startswith('_')
                and not key.startswith('_')
            )
        ]

        for name, func in operation_functions:
            title_name = ':code:`{0}.{1}`'.format(module_name, name)
            lines.append(title_name)

            # Underline name with -'s for title
            lines.append(_title_line('~', title_name))

            doc = func.__doc__
            if doc:
                docbits = doc.strip().split('\n')
                description_lines = []

                for line in docbits:
                    if line:
                        description_lines.append(line)
                    else:
                        break

                if len(docbits) > 0:
                    lines.append('')
                    lines.extend([line.strip() for line in description_lines])
                    lines.append('')
                    doc = '\n'.join(docbits[len(description_lines):])

            argspec = getargspec(func)

            # Make default strings appear as strings
            arg_defaults = [
                "'{}'".format(arg) if isinstance(arg, str) else arg
                for arg in argspec.defaults
            ] if argspec.defaults else None

            # Create a dict of arg name -> default
            defaults = dict(zip(
                argspec.args[-len(arg_defaults):],
                arg_defaults,
            )) if arg_defaults else {}

            # Build args string
            args = [
                '{0}={1}'.format(arg, defaults[arg])
                if arg in defaults else arg
                for arg in argspec.args[2:]
            ]

            if len(', '.join(args)) <= MODULE_DEF_LINE_MAX:
                args_string = ', '.join(args)

            else:
                arg_buffer = []
                arg_lines = []
                for arg in args:
                    if len(', '.join(arg_buffer + [arg])) >= MODULE_DEF_LINE_MAX:
                        arg_lines.append(arg_buffer)
                        arg_buffer = []

                    arg_buffer.append(arg)

                if arg_buffer:
                    arg_lines.append(arg_buffer)

                arg_lines = [
                    '        {0}'.format(', '.join(line_args))
                    for line_args in arg_lines
                ]

                args_string = '\n{0}\n    '.format(',\n'.join(arg_lines))

            # Atttach the code block
            lines.append('''
.. code:: python

    {0}.{1}({2})

'''.strip().format(module_name, name, args_string))

            # Append any remaining docstring
            if doc:
                lines.append('')
                lines.append('{0}'.format('\n'.join([
                    _format_doc_line(line) for line in doc.split('\n')
                ])).strip())

            lines.append('')
            lines.append('')

        # Write out the file
        module_filename = 'docs/modules/{0}.rst'.format(module_name)
        print('--> Writing {0}'.format(module_filename))

        out = '\n'.join(lines)

        outfile = open(module_filename, 'w')
        outfile.write(out)
        outfile.close()


if __name__ == '__main__':
    print('### Building module docs')
    build_facts()



# piynfra
# File: example/api_deploy.py
# Desc: example of how to deploy via the pyinfra API

from gevent import monkey  # noqa
monkey.patch_all()  # noqa async things (speed++, optional)

import json
import logging

from pyinfra.api import Config, Inventory, State
from pyinfra.api.connect import connect_all
from pyinfra.api.facts import get_facts
from pyinfra.api.operation import add_op
from pyinfra.api.operations import run_ops

from pyinfra.modules import files, server


# Enable pyinfra logging
logging.basicConfig(level=logging.WARNING)
logging.getLogger('pyinfra').setLevel(logging.INFO)


# First we setup some inventory we want to target
# the first argument is a tuple of (list all all hosts, global/ALL data)
inventory = Inventory(
    ([
        'centos6.pyinfra',
        # Host-specific data can be attached in inventory
        ('centos7.pyinfra', {'systemd': True}),
        'ubuntu14.pyinfra',
        'debian7.pyinfra',
        'openbsd58.pyinfra',
    ], {}),
    bsd=([
        'openbsd57.pyinfra',
    ], {
        # Group-specific data can be attached like so
        'app_dir': '/opt/pyinfra/bsd',
    }),
    centos=([
        'centos6.pyinfra',
        'centos7.pyinfra',
    ], {}),
    ssh_user='vagrant',
    ssh_key='./files/insecure_private_key',
)

# Now we create a new config (w/optional args)
config = Config(
    FAIL_PERCENT=81,
    TIMEOUT=5,
)

# Setup the pyinfra state for this deploy
state = State(inventory, config)


# Connect to all the hosts
print('Connecting...')
connect_all(state)


# Start adding operations
print('Generating operations...')
add_op(
    state, server.user,
    'pyinfra',
    home='/home/pyinfra',
    shell='/bin/bash',
    sudo=True,
)

add_op(
    state, server.group,
    {'Ensure pyinfra2 group exists'},  # set as the first arg names the operation
    'pyinfra2',
    sudo=True,
    # Add an op only to a subset of hosts
    # (in this case, the inventory.centos group)
    hosts=inventory.get_group('centos'),
)

# Ensure the state of files
add_op(
    state, files.file,
    '/var/log/pyinfra.log',
    user='pyinfra',
    group='pyinfra',
    mode='644',
    sudo=True,
)

# Ensure the state of directories
add_op(
    state, files.directory,
    '/tmp/email',
    user='pyinfra',
    group='pyinfra',
    mode='755',
    sudo=True,
)

# Copy local files to remote host
add_op(
    state, files.put,
    'files/file.txt', '/home/vagrant/file.txt',
)


# And finally we run the ops
run_ops(state)


# We can also get facts for all the hosts
facts = get_facts(state, 'os')
print(json.dumps(facts, indent=4))



# pyinfra
# File: pyinfra/example/deploy.py
# Desc: example deploy script for the pyinfra CLI
#       targets: Ubuntu/Debian, CentOS/Fedora & OpenBSD

# Host represents the *current* server begin managed
from pyinfra import hook, host, inventory, local

# Modules provide namespaced operations, which do the work
from pyinfra.modules import apt, files, git, init, pip, python, server, yum


# Hooks inside deploy file
@hook.before_connect
def before_connect(data, state):
    print('inventory hosts!: ', [host.name for host in inventory])


def on_pyinfra_success(state, host, op_hash):
    print('Success on {0} for OP: {1}!'.format(host.name, op_hash))


def on_pyinfra_error(state, host, op_hash):
    print('Error on {0} for OP: {1}!'.format(host.name, op_hash))


# Ensure the state of a user
server.user(
    'pyinfra',
    shell='/bin/sh',
    ensure_home=True,

    # Options available for all operations
    sudo=True,
    sudo_user='root',
    ignore_errors=False,
    serial=False,
    run_once=False,
    get_pty=False,
    timeout=30,  # only applies to commands on the remote host (not SFTP, local Python)
    on_success=on_pyinfra_success,
)

# And groups
server.group(
    {'Ensure pyinfra2 group exists'},  # use a set as the first arg to set the op name
    'pyinfra2',
    sudo=True,
    run_once=True,  # run only on one host
)

# Ensure the state of files
files.file(
    '/var/log/pyinfra.log',
    user='pyinfra',
    group='pyinfra',
    mode=644,
    sudo=True,
)

# Ensure the state of directories
files.directory(
    host.data.env_dir,
    user='pyinfra',
    group='pyinfra',
    mode=755,
    recursive=True,
    sudo=True,
    serial=True,
)

# Copy local files to remote host
files.put(
    'files/file.txt',
    '/home/vagrant/file.txt',
    mode=777,
)
# and sync directories
files.sync(
    'files',
    '/home/vagrant/example_files',
    user='pyinfra',
    group='pyinfra',
    delete=True,
    sudo=True,
)

# Generate files from local jinja2 templates
files.template(
    'templates/template.txt.j2',
    '/home/vagrant/template.txt',
)

# Execute arbitrary shell commands
server.shell([
    'echo "Shell command"',
    'echo "My hostname is {{ host.fact.hostname }}"',
])
# and scripts
server.script(
    'files/test.sh',
)

# Manage init systems
init.service(
    'cron',
    running=True,
    sudo=True,
    ignore_errors=True,
    on_error=on_pyinfra_error,
)

# Include roles
local.include(
    'roles/bsd_role.py',
    hosts=inventory.get_group('bsd'),
)

# Storing this fact to avoid typing it so much (because the example targets a whole bunch
# of distros [& 2 OSs]).
distro = host.fact.linux_distribution

# Work with facts about the remote host
if distro['name'] in ('Debian', 'Ubuntu'):
    # apt package manager
    apt.packages(
        ['git', 'python-pip'],
        sudo=True,
        update=True,
        cache_time=3600,
    )

elif distro['name'] in ('CentOS', 'Fedora'):
    if distro['name'] == 'CentOS':
        # Both missing in the CentOS 7 Vagrant image
        yum.packages(
            ['wget', 'net-tools'],
            sudo=True,
        )

        # Manage remote rpm files
        yum.rpm((
            'https://dl.fedoraproject.org/pub/epel/epel-release-latest-'
            '{{ host.fact.linux_distribution.major }}.noarch.rpm'
        ), sudo=True)

    # yum package manager
    yum.packages(
        ['git', 'python-pip'],
        sudo=True,
    )

    # Edit lines in files
    files.line(
        '/etc/sysconfig/selinux',
        '^SELINUX=.*',
        replace='SELINUX=disabled',
        sudo=True,
    )

# Ensure the state of git repositories
git.repo(
    'git@github.com:Fizzadar/pyinfra',
    host.data.app_dir,
    branch='develop',
    ssh_keyscan=True,
    run_once=True,
    sudo=True,
    # Carry SSH agent details w/sudo
    preserve_sudo_env=True,
)

# Manage pip (npm, gem) packages
did_install = pip.packages(
    ['virtualenv'],
    sudo=True,
)
# use operation meta to affect the deploy
if did_install.changed:
    server.shell(
        'echo "Clean package build/etc"',
    )

# Create a virtualenv
server.shell(
    'virtualenv {{ host.data.env_dir }}',
    sudo=True,
    sudo_user='pyinfra',
)
# and manage pip within it
pip.packages(
    ['ElasticQuery', 'JsonTest'],
    virtualenv=host.data.env_dir,
    sudo=True,
    sudo_user='pyinfra',
)

# Wait for services
server.wait(
    port=22,
    timeout=5,
)


# Execute Python locally, mid-deploy
def some_python(state, host, *args, **kwargs):
    print('connecting host name: {0}, actual: {1}'.format(host.name, host.fact.hostname))
    local.shell('echo "local stuff!"')


python.call(some_python, 'arg1', 'arg2', kwarg='hello world')



#!/usr/bin/env python

# pyinfra
# File: docs/generate_fact_docs.py
# Desc: generate rst docs from the fact classes

from inspect import getargspec, getmembers, isclass
from types import MethodType

from six.moves import range

from pyinfra import facts
from pyinfra.api.facts import FactBase
from pyinfra.api.util import underscore


def _title_line(char, string):
    return ''.join(char for _ in range(0, len(string)))


def build_facts():
    lines = []

    # Now we generate a facts.rst describing the use of the facts as:
    # host.data.<snake_case_fact>
    for module_name in facts.__all__:
        print('--> Doing fact module: {0}'.format(module_name))
        module = getattr(facts, module_name)

        lines.append(module_name.title())
        lines.append(_title_line('-', module_name))
        lines.append('')

        fact_classes = [
            (key, value)
            for key, value in getmembers(module)
            if (
                isclass(value)
                and issubclass(value, FactBase)
                and value.__module__ == module.__name__
                and value is not FactBase
            )
        ]

        for fact, cls in fact_classes:
            # FactClass -> fact_accessor on host object
            name = underscore(fact)

            # Does this fact take args?
            command_attr = getattr(cls, 'command', None)
            if isinstance(command_attr, MethodType):
                # Attach basic argspec to name
                # Note only supports facts with one arg as this is all that's possible,
                # will need to refactor to print properly in future.
                argspec = getargspec(command_attr)
                if len(argspec.args) > 1:
                    name = '{0}({1})'.format(
                        name,
                        ', '.join(arg for arg in argspec.args if arg != 'self'),
                    )

            name = ':code:`{0}`'.format(name)
            lines.append(name)

            # Underline name with -'s for title
            lines.append(_title_line('~', name))

            # Append any docstring
            doc = cls.__doc__
            if doc:
                lines.append('')
                lines.append('{0}'.format('\n'.join([
                    line for line in doc.split('\n')
                ])))

            lines.append('')
            lines.append('')

    # Write out the file
    print('--> Writing docs/facts.rst')

    out = '\n'.join(lines)
    out = '''
Facts Index
===========

.. include:: facts_.rst


{0}
    '''.format(out).strip()

    outfile = open('docs/facts.rst', 'w')
    outfile.write(out)
    outfile.close()


if __name__ == '__main__':
    print('### Building fact docs')
    build_facts()



# pyinfra
# File: example/roles/bsd_role.py
# Desc: example role that installs pip on (Open)BSD

from pyinfra.modules import server, pkg


# OpenBSD packages?
pkg.packages(
    ['py-pip', 'git'],
    sudo=True
)

# add_pkg does not automatically do this
server.shell(
    'ln -sf /usr/local/bin/pip2.7 /usr/local/bin/pip',
    sudo=True
)



# pyinfra
# File: pyinfra/example/config.py
# Desc: entirely optional config file for the CLI deploy
#       see: pyinfra/api/config.py for defaults

from pyinfra import local, hook

# These can be here or in deploy.py
TIMEOUT = 5
FAIL_PERCENT = 81


# Deploy hook, can also be in deploy.py
@hook.before_connect
def ensure_branch(data, state):
    # Check something local is correct, etc
    branch = local.shell('git rev-parse --abbrev-ref HEAD')
    app_branch = data.app_branch

    if branch != app_branch:
        # Raise hook.Error for pyinfra to handle
        raise hook.Error('We\'re on the wrong branch (want {0}, got {1})!'.format(
            app_branch, branch
        ))


@hook.after_deploy
def notify_people(data, state):
    print('After deploy hook!')



# pyinfra
# File: example/inventories/dev.py
# Desc: hostname based inventory (requires ./hosts in /etc/hosts)

# Defines a group
bsd = [
    'openbsd58.pyinfra',
]

yum = [
    'centos6.pyinfra',
    'centos7.pyinfra',
    'fedora23.pyinfra'
]

apt = [
    # Host-specific data can be attached in inventory
    ('ubuntu14-hosttest', {'ssh_hostname': 'ubuntu14.pyinfra'}),
    'ubuntu15.pyinfra',
    'debian7.pyinfra',
    'debian8.pyinfra'
]

# Hosts can be in multiple groups
linux = yum + apt

# The all group is automatically set to the unique hosts above and so does not need to be defined
# all = linux + bsd



# pyinfra
# File: example/group_data/all.py
# Desc: config/variables which apply to all hosts, must be snake_case

# Required
ssh_user = 'vagrant'
ssh_key = 'files/insecure_private_key'

# Anything else
env_dir = '/opt/env'
app_dir = '/opt/pyinfra'
app_branch = 'develop'
